import torch
import copy
import os
from tqdm import tqdm
from PIL import Image

import torch.utils.data as data
from pycocotools.coco import COCO
from gluoncv.utils.bbox import bbox_xywh_to_xyxy, bbox_clip_xyxy

from detectron2.structures import Instances, Boxes

class ImageDataset(data.Dataset):
    def __init__(self,
                 dataset: COCO,
                 image_dir,
                 transforms=None,
                 frames_per_image=1,
                 amodal=False,
                 skip_empty=True,
                 min_object_area=0,
                 use_crowd=False,
                 include_bg=False,
                 ):
        """
        :param dataset: the ingested dataset with COCO-format
        :param transforms: image transformation
        :param frames_per_image: how many image copies are generated from a single image
        :param amodal: whether to use amodal ground truth (no image boundary clipping)
        :param include_bg: whether to include the full background images during training
        """

        self.dataset = dataset
        self.image_dir = image_dir
        self.transforms = transforms
        self.frames_per_image = frames_per_image

        self._skip_empty = skip_empty
        self._min_object_area = min_object_area
        self._use_crowd = use_crowd
        self._amodal = amodal
        self._include_bg = include_bg
        self._det_classes = [c['name'] for c in self.dataset.loadCats(self.dataset.getCatIds())]

        # These are tha mapping table of COCO labels
        self.json_category_id_to_contiguous_id = {
            v: i for i, v in enumerate(self.dataset.getCatIds())
        }

        self._labels, self._im_aspect_ratios, self._items, self._ids \
            = self._dataset_preprocess()

        self.id_to_img_map = {k: v for k, v in enumerate(self._ids)}

    def __getitem__(self, index):
        img_name = self._items[index]
        img_path = os.path.join(self.image_dir, img_name)

        img = Image.open(img_path).convert('RGB')
        w, h = img.size
        target = self._get_target(img, index)

        # for tracking purposes, two frames are needed
        # the pairs would go into random augmentation to generate fake motion
        video_clip = [img for _ in range(self.frames_per_image)]
        video_target = [target for _ in range(self.frames_per_image)]

        if self.transforms is not None:
            video_clip, video_target = self.transforms(video_clip, video_target)

        video_target_ = []
        TO_REMOVE = 1
        vt = copy.deepcopy(video_target)
        for i, vt_ in enumerate(vt):
            h, w = vt_.image_size
            vt[i].gt_boxes.tensor[:, 0].clamp_(min=0, max=w - TO_REMOVE)
            vt[i].gt_boxes.tensor[:, 1].clamp_(min=0, max=h - TO_REMOVE)
            vt[i].gt_boxes.tensor[:, 2].clamp_(min=0, max=w - TO_REMOVE)
            vt[i].gt_boxes.tensor[:, 3].clamp_(min=0, max=h - TO_REMOVE)
            box = vt[i].gt_boxes.tensor
            keep = (box[:, 3] > box[:, 1]) & (box[:, 2] > box[:, 0])
            video_target_.append(vt[i][keep])

        outputs = []
        for idx in range(len(video_clip)):
            outputs.append({
                "image_id": index,
                "height": h,
                "width": w,
                "image": video_clip[idx],
                "instances": video_target_[idx]
            })

        return outputs

    def _get_target(self, img, index):
        labels = self._labels[index]
        if len(labels) == 0:
            assert self._include_bg is True, "The image does not has ground truth"
            bbox = torch.as_tensor(labels).reshape(-1, 4)
            class_ids = torch.as_tensor(labels)
            instance_ids = torch.as_tensor(labels)

            w, h = img.size
            bbox = Boxes(bbox)
            instance = Instances((h, w))
            instance.set("gt_boxes", bbox)
            instance.set("gt_classes", class_ids)
            instance.set("ids", instance_ids)
            return instance

        labels = torch.as_tensor(labels).reshape(-1, 6)
        boxes = labels[:, :4]

        w, h = img.size
        bbox = Boxes(boxes)

        instance = Instances((h, w))
        instance.set("gt_boxes", bbox)

        class_ids = labels[:, 4].clone().to(torch.int64)
        instance.set("gt_classes", class_ids)

        instance_ids = labels[:, -1].clone().to(torch.int64)
        instance.set("ids", instance_ids)
        return instance

    def _dataset_preprocess(self):
        items = []
        labels = []
        ids = []
        im_aspect_ratios = []
        image_ids = sorted(self.dataset.getImgIds())
        instance_id = 0
        rm_redundant = 0
        all_amodal = 0

        for entry in tqdm(self.dataset.loadImgs(image_ids)):
            label, num_instances, num_redundant, num_amodal\
                = self._check_load_bbox(entry, instance_id)
            if not label and not self._include_bg:
                continue
            instance_id += num_instances
            rm_redundant += num_redundant
            all_amodal += num_amodal
            labels.append(label)
            ids.append(entry['id'])
            items.append(entry['file_name'])
            im_aspect_ratios.append(float(entry['width']) / entry['height'])

        print('{} / {} valid images...'.format(len(labels), len(image_ids)))
        print('{} instances...'.format(instance_id))
        print('{} redundant instances are removed...'.format(rm_redundant))
        print('{} amodal instances...'.format(all_amodal))
        return labels, im_aspect_ratios, items, ids

    def _check_load_bbox(self, entry, instance_id):
        """
        Check and load ground-truth labels
        """
        entry_id = entry['id']
        entry_id = [entry_id] if not isinstance(entry_id, (list, tuple)) else entry_id
        ann_ids = self.dataset.getAnnIds(imgIds=entry_id, iscrowd=None)
        objs = self.dataset.loadAnns(ann_ids)

        # check valid bboxes
        valid_objs = []
        width = entry['width']
        height = entry['height']
        _instance_count = 0
        _redudant_count = 0
        _amodal_count = 0
        unique_bbs = set()
        for obj in objs:
            if obj.get('ignore', 0) == 1:
                continue
            if not self._use_crowd and obj.get('iscrowd', 0):
                continue
            if self._amodal:
                xmin, ymin, xmax, ymax = bbox_xywh_to_xyxy(obj['bbox'])
                if xmin < 0 or ymin < 0 or xmax > width or ymax > height:
                    _amodal_count += 1
            else:
                xmin, ymin, xmax, ymax = bbox_clip_xyxy(bbox_xywh_to_xyxy(obj['bbox']), width, height)

            if (xmin, ymin, xmax, ymax) in unique_bbs:
                _redudant_count += 1
                continue

            box_w = (xmax - xmin)
            box_h = (ymax - ymin)
            area = box_w * box_h
            if area <= self._min_object_area:
                continue

            # require non-zero box area
            if xmax > xmin and ymax > ymin:
                unique_bbs.add((xmin, ymin, xmax, ymax))
                contiguous_cid = self.json_category_id_to_contiguous_id[obj['category_id']]
                valid_objs.append([xmin, ymin, xmax, ymax, contiguous_cid,
                                   instance_id+_instance_count])
                _instance_count += 1
        if not valid_objs:
            if not self._skip_empty:
                # dummy invalid labels if no valid objects are found
                valid_objs.append([-1, -1, -1, -1, -1, -1])
        return valid_objs, _instance_count, _redudant_count, _amodal_count

    def __len__(self):
        return len(self._items)

    def get_img_info(self, index):
        img_id = self.id_to_img_map[index]
        img_data = self.dataset.imgs[img_id]
        return img_data

    @property
    def classes(self):
        return self._det_classes

    def get_im_aspect_ratio(self):
        return self._im_aspect_ratios